/*
    rtl/com.h

       ,  COM-.
*/
#ifndef _rtl_com_h_included_
#define _rtl_com_h_included_

#include <windows.h>
#include <objbase.h>
#include <ocidl.h>

#ifdef __cplusplus
extern "C" {
#endif

/* -------------------------------------------------------------------------- */
/*  */

/*
      callback-  COM_EnumXXX
*/
enum COM_ENUM_RESULT
{
    COM_ENUM_CONTINUE,      /*    */
    COM_ENUM_SKIP_CLASS,    /*           */
    COM_ENUM_BREAK          /*    */
};

/*
     
*/
enum COM_ERROR_CODES
{
    E_NOTFOUND      = __HRESULT_FROM_WIN32 (ERROR_NOT_FOUND),
    E_ALREADYEXISTS = __HRESULT_FROM_WIN32 (ERROR_ALREADY_EXISTS),
    E_INVALIDTYPE   = __HRESULT_FROM_WIN32 (ERROR_INVALID_DATATYPE),
    E_EMPTY         = __HRESULT_FROM_WIN32 (ERROR_NO_MORE_ITEMS),
};

/* -------------------------------------------------------------------------- */
/*  */

/*
         / 
*/
#define COM_CLSID(className)                    CLSID_##className
#define COM_IID(intfName)                       IID_##intfName
#define COM_VTABLE_NAME(className, intfName)    className##_##intfName##Vtbl
#define COM_INTF_MAP_NAME(className)            className##_InterfaceMap
#define COM_CLASS_NAME(className)               className##_Class
#define COM_RUNTIME_CLASS_NAME(className)       className##_RuntimeClass
#define COM_CTOR_NAME(className)                className##_InitInstance
#define COM_DTOR_NAME(className)                className##_FreeInstance

#ifdef COM_NO_RTTI_NAMES
#   define COM_NAME_STRING(x)   NULL
#else
#   define COM_NAME_STRING(x)   #x
#endif

#define COM_MAKE_VERSION(major, minor)  ((DWORD)(((((DWORD)(major)) & 0xffff) << 16) | (((DWORD)(minor)) & 0xffff)))
#define COM_MAJOR_VERSION(version)      ((WORD)((((DWORD)(version)) >> 16) & 0xffff))
#define COM_MINOR_VERSION(version)      ((WORD)(((DWORD)(version)) & 0xffff))

#define COM_ANY_VERSION                 ((WORD)0xffff)

#define COM_FIELD_OFFSET(_class, field) ((DWORD)(DWORD_PTR)&(((_class*)0)->field))
#define COM_FIELD_SIZE(_class, field)   ((DWORD)(DWORD_PTR)sizeof(((_class*)0)->field))

/*
    COM_DECLARE_VTABLE

        .

    :

    COM_DECLARE_VTABLE (CSomeClass, ISomeInterface)
    {
        // IUnknown
        COM_IUNKNOWN_METHODS (ISomeInterface),

        // ISomeInterface
        CSomeClass_Method1,
        CSomeClass_Method2,
        ...
    };
*/
#define COM_DECLARE_VTABLE(className, intfName)  \
    static CONST_VTBL intfName##Vtbl COM_VTABLE_NAME(className, intfName) =

/*
    COM_INIT_PTR_ARG

     [out]     E_POINTER,  
      NULL.
*/
#define COM_INIT_PTR_ARG_(name, value) \
    if (name) \
        *name = value; \
    else \
        return E_POINTER

#define COM_INIT_PTR_ARG(name) \
    COM_INIT_PTR_ARG_ (name, NULL)

/*
    COM_METHOD_PROLOGUE(object, condition)

        object     This  
     E_INVALIDARG,   condition .

     condition    .
*/
#define COM_METHOD_PROLOGUE(object, ...) \
    ((COM_OBJECT*)object) = _COM_GetObject (This); \
    if (__VA_ARGS__) \
        return E_INVALIDARG

/*
    :

    STDMETHODIMP CObject_Method (IObject* This, BSTR bstrArg, BSTR* pbstrArg, DWORD* pdwArg)
    {
        CObject* object;

        COM_INIT_PTR_ARG (pbstrArg);
        COM_INIT_PTR_ARG_ (pdwArg, 0);
        COM_METHOD_PROLOGUE (object, !object || !bstrArg);

        return S_OK;
    }
*/

/* -------------------------------------------------------------------------- */
/*  */

typedef struct _COM_INTERFACE       COM_INTERFACE;
typedef struct _COM_CLASS           COM_CLASS;
typedef struct _COM_RUNTIME_CLASS   COM_RUNTIME_CLASS;
typedef struct _COM_OBJECT          COM_OBJECT;
typedef struct _COM_EVENTS          COM_EVENTS;
typedef struct _COM_STATS           COM_STATS;

typedef struct _COM_LIST_ITEM       COM_LIST_ITEM;
typedef struct _COM_LIST_ITER       COM_LIST_ITER;
typedef struct _COM_LIST            COM_LIST;

typedef struct _COM_TREE_NODE       COM_TREE_NODE;
typedef struct _COM_TREE_ITER       COM_TREE_ITER;
typedef struct _COM_TREE            COM_TREE;

/* -------------------------------------------------------------------------- */
/*
          AVL-.
*/

struct _COM_LIST_ITEM
{
    COM_LIST_ITEM* prev;
    COM_LIST_ITEM* next;

    /*   */
};

struct _COM_LIST_ITER
{
    COM_LIST_ITEM iterItem;

    COM_LIST_ITEM* item;
};

#define COM_LIST_ITER_FIRST_ITEM    ((COM_LIST_ITEM*)1)

struct _COM_LIST
{
    COM_LIST_ITEM* head;
    COM_LIST_ITEM* tail;
    COM_LIST* iterators;

    void (* Free)(COM_LIST* list, COM_LIST_ITEM* item);
};

struct _COM_TREE_NODE
{
    COM_TREE_NODE* parent;
    COM_TREE_NODE* left;
    COM_TREE_NODE* right;
    int bf;

    /*   */
};

struct _COM_TREE_ITER
{
    COM_LIST_ITEM item;

    COM_TREE_NODE* node;
};

#define COM_TREE_ITER_FIRST_NODE    ((COM_TREE_NODE*)1)

struct _COM_TREE
{
    COM_TREE_NODE* root;
    COM_LIST* iterators;

    void    (* Free)    (COM_TREE* tree, COM_TREE_NODE* node);
    int     (* Compare) (COM_TREE* tree, LPCVOID key, COM_TREE_NODE* node);
};

/* -------------------------------------------------------------------------- */
/*
     

       -    COM_INTERFACE,  
    ,   .

        BEGIN_INTERFACE_MAP(CClassName)  
      END_INTERFACE_MAP(),   
    COM_INTERFACE_ENTRY*()   ,  .

       ,   :
        -  CClassName,
        -  IID_IInterfaceName      
          .
        -     CClassName_IInterfaceNameVtbl  
           ,   .
*/

enum COM_INTERFACE_TYPE
{
    CIT_END,
    CIT_GENERIC,
    CIT_AGGREGATE,
    CIT_AGGREGATE2,
    CIT_NOINTERFACE,
    CIT_CALLBACK,

    CIT_LAST
};

struct _COM_INTERFACE
{
    int type;
    const IID* iid;
    LPCSTR intfName;
    DWORD fieldOffset;
    DWORD fieldSize;
    LPCSTR fieldName;
    LPCVOID lpVtbl;

    union
    {
        LPCVOID ptr;
        const CLSID* innerCLSID;
        const COM_CLASS* innerClass;
        const GUID* libID;
        HRESULT (*QI)(COM_OBJECT* object, DWORD dw, REFIID iid, LPVOID* ppvObject);
    };

    union
    {
        DWORD libVersion;
        DWORD dw;
    };

    LCID libLCID;
};

#define BEGIN_INTERFACE_MAP(className) \
    static const COM_INTERFACE COM_INTF_MAP_NAME(className)[] = \
    {

#define COM_INTERFACE_ENTRY(intf, className, field)                                     {CIT_GENERIC, &COM_IID(intf), COM_NAME_STRING(intf), COM_FIELD_OFFSET(className, field), COM_FIELD_SIZE(className, field), COM_NAME_STRING(field), &COM_VTABLE_NAME(className, intf), NULL, 0, 0}
#define COM_INTERFACE_ENTRY2(intf, className, field, libid, version, lcid)              {CIT_GENERIC, &COM_IID(intf), COM_NAME_STRING(intf), COM_FIELD_OFFSET(className, field), COM_FIELD_SIZE(className, field), COM_NAME_STRING(field), &COM_VTABLE_NAME(className, intf), &(libid), (version), (lcid)}
#define COM_INTERFACE_ENTRY_AGGREGATE(intf, className, field, clsid)                    {CIT_AGGREGATE, &COM_IID(intf), COM_NAME_STRING(intf), COM_FIELD_OFFSET(className, field), COM_FIELD_SIZE(className, field), COM_NAME_STRING(field), NULL, &clsid, 0, 0}
#define COM_INTERFACE_ENTRY_AGGREGATE_BLIND(className, field, clsid)                    {CIT_AGGREGATE, NULL, NULL, COM_FIELD_OFFSET(className, field), COM_FIELD_SIZE(className, field), COM_NAME_STRING(field), NULL, &(clsid), 0, 0}
#define COM_INTERFACE_ENTRY_AGGREGATE2(intf, className, field, innerClass)              {CIT_AGGREGATE2, &COM_IID(intf), COM_NAME_STRING(intf), COM_FIELD_OFFSET(className, field), COM_FIELD_SIZE(className, field), COM_NAME_STRING(field), NULL, &COM_CLASS_NAME(innerClass), 0, 0}
#define COM_INTERFACE_ENTRY_AGGREGATE2_BLIND(className, field, innerClass)              {CIT_AGGREGATE2, NULL, NULL, COM_FIELD_OFFSET(className, field), COM_FIELD_SIZE(className, field), COM_NAME_STRING(field), NULL, &COM_CLASS_NAME(innerClass), 0, 0}
#define COM_INTERFACE_ENTRY_NOINTERFACE(intf)                                           {CIT_NOINTERFACE, &COM_IID(intf), COM_NAME_STRING(intf), 0, 0, NULL, NULL, NULL, 0, 0}
#define COM_INTERFACE_ENTRY_CALLBACK(callback, dw)                                      {CIT_CALLBACK, NULL, NULL, 0, 0, NULL, NULL, callback, dw, 0}

#define END_INTERFACE_MAP() \
        {CIT_END, NULL, NULL, 0, 0, NULL, NULL, NULL, 0, 0}  \
    };

#define COM_INTERFACE_MAP(className) \
    &(COM_INTF_MAP_NAME(className))[0]

/* -------------------------------------------------------------------------- */
/* 
    COM_RUNTIME_CLASS

      ,        
       .
    
       (COM_CLASS::runtimeClass   NULL),   
           .

       -   .

      COM_DECLARE_RUNTIME_CLASS().
*/

#define COM_FLAGS_SINGLETON         0x00000001  /* - */
#define COM_FLAGS_NO_AGGREGATION    0x00000002  /*       */
#define COM_FLAGS_ONLY_AGGREGATION  0x00000004  /*      */

struct _COM_RUNTIME_CLASS
{
    COM_LIST_ITEM listItem;

    DWORD flags;                    /*    */
    DWORD intfTableSize;            /*    */
    DWORD objectSize;               /*    */
    DWORD instanceCount;            /*     */
    IUnknown* classFactory;         /*     */
    COM_OBJECT* singletonInstance;  /*  - */
    const COM_CLASS* _class;        /*     */
    CRITICAL_SECTION* lock;         /*    */
    COM_LIST objectList;            /*     */
};

#define COM_DECLARE_RUNTIME_CLASS(className, flags)                 \
    static COM_RUNTIME_CLASS COM_RUNTIME_CLASS_NAME(className) =    \
    {                                                               \
        /* item                 */ {NULL, NULL},                    \
        /* flags                */ flags,                           \
        /* intfTableSize        */ (DWORD)-1,                       \
        /* objectSize           */ 0,                               \
        /* instanceCount        */ 0,                               \
        /* classFactory         */ NULL,                            \
        /* singletonInstance    */ NULL,                            \
        /* _class               */ NULL,                            \
        /* lock                 */ NULL,                            \
        /* objectList           */ {NULL, NULL, NULL},              \
    }

/* -------------------------------------------------------------------------- */
/*
    COM_CLASS

      ,    
      .

      COM_DECLARE_CLASS().
*/

struct _COM_CLASS
{
    const COM_CLASS* parent;

    const CLSID* id;
    LPCSTR name;

    DWORD instanceSize;

    const COM_RUNTIME_CLASS* runtimeClass;

    HRESULT     (* InitInstance ) (COM_OBJECT* object);
    void        (* FreeInstance ) (COM_OBJECT* object);

    const COM_INTERFACE* interfaces;
};

/* -------------------------------------------------------------------------- */
/*
    COM_DECLARE_EXTERN_CLASS(className)

    COM_DECLARE_PUBLIC_CLASS  (parent, className, flags)
    COM_DECLARE_PUBLIC_CLASS_CLSID  (parent, clsid, className, flags)
    COM_DECLARE_PUBLIC_CLASS_NAME  (parent, name, className, flags)
    COM_DECLARE_PUBLIC_CLASS_CLSID_NAME (parent, clsid, name, className, flags)

    COM_DECLARE_PRIVATE_CLASS  (parent, className, flags)
    COM_DECLARE_PRIVATE_CLSID  (parent, clsid, className, flags)
    COM_DECLARE_PRIVATE_NAME  (parent, name, className, flags)
    COM_DECLARE_PRIVATE_CLASS_CLSID_NAME (parent, clsid, name, className, flags)

    _COM_DECLARE_CLASS (scope, parent, clsid, name, className, flags)

         .

         :
        -    className,
        -  className_InitInstance  className_FreeInstance,
        -     className.

    :
        parent      -      NULL
        clsid       - GUID 
        name        -    ,      
                       ,  "CTemplateClass<COM_CLASS>"
        className   -   ( ),  CSomeClass
        flags       -   COM_FLAGS_*
*/

#define _COM_DECLARE_CLASS(scope, parent, clsid, name, className, flags)    \
    COM_DECLARE_RUNTIME_CLASS(className, flags);                            \
    scope COM_CLASS COM_CLASS_NAME(className) =                             \
    {                                                                       \
        parent,                                                             \
        &clsid,                                                             \
        name,                                                               \
        sizeof(className),                                                  \
        &COM_RUNTIME_CLASS_NAME(className),                                 \
        COM_CTOR_NAME(className),                                           \
        COM_DTOR_NAME(className),                                           \
        COM_INTERFACE_MAP(className)                                        \
    }

#define COM_DECLARE_EXTERN_CLASS(className) \
    extern const COM_CLASS COM_CLASS_NAME(className)

#define COM_DECLARE_PUBLIC_CLASS(parent, className, flags) \
    _COM_DECLARE_CLASS (const, parent, COM_CLSID(className), COM_NAME_STRING(className), className, flags)

#define COM_DECLARE_PUBLIC_CLASS_CLSID(parent, clsid, className, flags) \
    _COM_DECLARE_CLASS (const, parent, clsid, COM_NAME_STRING(className), className, flags)

#define COM_DECLARE_PUBLIC_CLASS_NAME(parent, name, className, flags) \
    _COM_DECLARE_CLASS (const, parent, COM_CLSID(className), name, className, flags)

#define COM_DECLARE_PUBLIC_CLASS_CLSID_NAME(parent, clsid, name, className, flags) \
    _COM_DECLARE_CLASS (const, parent, clsid, name, className, flags)

#define COM_DECLARE_PRIVATE_CLASS(parent, className, flags) \
    _COM_DECLARE_CLASS (static const, parent, COM_CLSID(className), COM_NAME_STRING(className), className, flags)

#define COM_DECLARE_PRIVATE_CLASS_CLSID(parent, clsid, className, flags) \
    _COM_DECLARE_CLASS (static const, parent, clsid, COM_NAME_STRING(className), className, flags)

#define COM_DECLARE_PRIVATE_CLASS_NAME(parent, name, className, flags) \
    _COM_DECLARE_CLASS (static const, parent, COM_CLSID(className), name, className, flags)

#define COM_DECLARE_PRIVATE_CLASS_CLSID_NAME(parent, clsid, name, className, flags) \
    _COM_DECLARE_CLASS (static const, parent, clsid, name, className, flags)

/* -------------------------------------------------------------------------- */
/*
    COM_OBJECT

         .

      COM_OBJECT     .
*/

struct _COM_OBJECT
{
    IUnknown unknInner;
    IUnknown* unknOuter;
    volatile DWORD refCount;

    const COM_CLASS* _class;

    COM_LIST_ITEM listItem;
};

/* -------------------------------------------------------------------------- */
/*
    COM_EVENTS

       callback-,    
       -  .

         NULL.

        COM_GetEvents()    
    COM_Initialize()    COM_Cleanup(). 

       ,    COM_EVENTS  
        COM_EnterEvents()/COM_LeaveEvents().

        -   ,  
     COM_SetEventsParam().
*/

struct _COM_EVENTS
{
    /* 
             .
        
        message -    .
    */
    void (* OnErrorMessage)(LPCSTR message, LPVOID param);

    /* 
          .

        context - ,  .   NULL.
        name    -  .
    */
    void (* OnIntegerOverflow)(LPCSTR context, LPCSTR name, LPVOID param);

    /* 
                .
        
        context - ,  .   NULL.
        name    -  .
    */
    void (* OnNegativeCount)(LPCSTR context, LPCSTR name, LPVOID param);

    /*
            .

    */
    void (* OnLastObjectReleased)(LPVOID param);

    /*
           COM_Cleanup(),   
        .

              COM_EnumXXX().
    */
    void (* OnMemoryLeak)(LPVOID param);
};

/* -------------------------------------------------------------------------- */
/*
    COM_STATS

        .
*/

struct _COM_STATS
{
    BOOL isActive;          /*     */

    SYSTEMTIME startTime;   /* /   */
    SYSTEMTIME runTime;     /*    */

    DWORD objectCount;      /*     */
    DWORD lockCount;        /*     COM_LockServer() */

    UINT64 memoryUsed;      /*       */

    DWORD threadId;         /*  ,      */
};

/* -------------------------------------------------------------------------- */
/*   IUnknown   COM_CLASS */

HRESULT WINAPI  CUnknown_QueryInterface (IUnknown* This, REFIID riid, LPVOID* ppvObject);
ULONG WINAPI    CUnknown_AddRef         (IUnknown* This);
ULONG WINAPI    CUnknown_Release        (IUnknown* This);

#define COM_IUNKNOWN_METHODS(iface) \
    (HRESULT (STDMETHODCALLTYPE*)(iface*, REFIID, LPVOID*))CUnknown_QueryInterface, \
    (ULONG (STDMETHODCALLTYPE*)(iface*))CUnknown_AddRef, \
    (ULONG (STDMETHODCALLTYPE*)(iface*))CUnknown_Release

/* -------------------------------------------------------------------------- */
/*   IDispatch   COM_CLASS */

HRESULT WINAPI  CDispatch_GetTypeInfoCount  (IDispatch* This, UINT* pctinfo);
HRESULT WINAPI  CDispatch_GetTypeInfo       (IDispatch* This, UINT iTInfo, LCID lcid, ITypeInfo** ppTInfo);
HRESULT WINAPI  CDispatch_GetIDsOfNames     (IDispatch* This, REFIID riid, LPOLESTR* rgszNames, UINT cNames, LCID lcid, DISPID* rgDispId);
HRESULT WINAPI  CDispatch_Invoke            (IDispatch* This, DISPID dispIdMember, REFIID riid, LCID lcid, WORD wFlags, DISPPARAMS* pDispParams, VARIANT* pVarResult, EXCEPINFO* pExcepInfo, UINT* puArgErr);

#define COM_IDISPATCH_METHODS(iface) \
    (HRESULT (STDMETHODCALLTYPE*)(iface*, UINT*))CDispatch_GetTypeInfoCount, \
    (HRESULT (STDMETHODCALLTYPE*)(iface*, UINT, LCID, ITypeInfo**))CDispatch_GetTypeInfo, \
    (HRESULT (STDMETHODCALLTYPE*)(iface*, REFIID, LPOLESTR*, UINT, LCID, DISPID*))CDispatch_GetIDsOfNames, \
    (HRESULT (STDMETHODCALLTYPE*)(iface*, DISPID, REFIID, LCID, WORD, DISPPARAMS*, VARIANT*, EXCEPINFO*, UINT*))CDispatch_Invoke

/* -------------------------------------------------------------------------- */
/*                                                          */
/* -------------------------------------------------------------------------- */
/*
    IGuid

      GUID.

            COM_OBJECT.  QueryInterface
      .

      GUID   .
*/

typedef struct _IGuidVtbl   IGuidVtbl;
typedef struct _IGuid       IGuid;

struct _IGuidVtbl
{
    /* IUnknown */
    HRESULT (WINAPI *QueryInterface)    (IGuid* This, REFGUID iid, LPVOID* ppvObject);
    ULONG   (WINAPI *AddRef)            (IGuid* This);
    ULONG   (WINAPI *Release)           (IGuid* This);

    /* IGuid */
    REFGUID (WINAPI *Value)             (IGuid* This);
    int     (WINAPI *Compare)           (IGuid* This, REFGUID guid);
    BOOL    (WINAPI *Equal)             (IGuid* This, REFGUID guid);
};

struct _IGuid
{
    const IGuidVtbl* lpVtbl;
};

#ifdef COBJMACROS
#   define IGuid_QueryInterface(This, iid, ppvObject)   ((This)->lpVtbl->QueryInterface (This, iid, ppvObject))
#   define IGuid_AddRef(This)                           ((This)->lpVtbl->AddRef (This))
#   define IGuid_Release(This)                          ((This)->lpVtbl->Release (This))
#   define IGuid_Value(This)                            ((This)->lpVtbl->Value (This))
#   define IGuid_Compare(This, guid)                    ((This)->lpVtbl->Compare (This, guid))
#   define IGuid_Equal(This, guid)                      ((This)->lpVtbl->Equal (This, guid))
#endif

/* -------------------------------------------------------------------------- */
/* 
     IEnumerable

     ,      
     IEnum<T>.

    :

    CreateIterator (LPVOID* iter)

          ,     .

    CopyIterator (LPVOID iter, LPVOID* newIter)

          ,     ,   
        .

    FreeIterator (LPVOID iter)

         ,   CreateIterator()  CopyIterator().

    Next (LPVOID* iter, ULONG count, LPVOID buffer, ULONG* pFetched)

           count ,   ,    
           .  S_OK,   count
          S_FALSE,    .
         pFetched  NULL,      .

    Skip (LPVOID* iter, ULONG count)
        
           count  ("").
*/

/* {7C0D5B65-F7B7-4593-BA14-118D9C02F78C} */
static const GUID IID_IEnumerable = {0x7c0d5b65, 0xf7b7, 0x4593, {0xba, 0x14, 0x11, 0x8d, 0x9c, 0x2, 0xf7, 0x8c}};

typedef struct _IEnumerableVtbl IEnumerableVtbl;
typedef struct _IEnumerable     IEnumerable;

struct _IEnumerableVtbl
{
    /* IUnknown */
    STDMETHOD  (QueryInterface  ) (IEnumerable* This, REFIID iid, LPVOID* ppvObject);
    STDMETHOD_ (ULONG, AddRef   ) (IEnumerable* This);
    STDMETHOD_ (ULONG, Release  ) (IEnumerable* This);

    /* IEnumerable */
    STDMETHOD  (CreateIterator  ) (IEnumerable* This, LPVOID* iter);
    STDMETHOD  (CopyIterator    ) (IEnumerable* This, LPVOID iter, LPVOID* newIter);
    STDMETHOD  (FreeIterator    ) (IEnumerable* This, LPVOID iter);
    STDMETHOD  (Next            ) (IEnumerable* This, LPVOID* iter, ULONG count, LPVOID buffer, ULONG* pFetched/*=NULL*/);
    STDMETHOD  (Skip            ) (IEnumerable* This, LPVOID* iter, ULONG count);
};

struct _IEnumerable
{
    const IEnumerableVtbl* lpVtbl;
};

#ifdef COBJMACROS
#   define IEnumerable_QueryInterface(This, iid, ppvObject)         ((This)->lpVtbl->QueryInterface (This, iid, ppvObject))
#   define IEnumerable_AddRef(This)                                 ((This)->lpVtbl->AddRef (This))
#   define IEnumerable_Release(This)                                ((This)->lpVtbl->Release (This))
#   define IEnumerable_CreateIterator(This, iter)                   ((This)->lpVtbl->CreateIterator (This, iter))
#   define IEnumerable_CopyIterator(This, iter, newIter)            ((This)->lpVtbl->CopyIterator (This, iter, newIter))
#   define IEnumerable_FreeIterator(This, iter)                     ((This)->lpVtbl->FreeIterator (This, iter))
#   define IEnumerable_Next(This, iter, count, buffer, pFetched)    ((This)->lpVtbl->Next (This, iter, count, buffer, pFetched))
#   define IEnumerable_Skip(This, iter, count)                      ((This)->lpVtbl->Skip (This, iter, count))
#endif

/* -------------------------------------------------------------------------- */
/*
     IUnknArray

     ,   IUnknown.

    :

    SetMaxSize (DWORD count)
        
             .  
          0     .

    Add (IUnknown* unkn, REFIID iid, DWORD* pCookie)

            .  iid   
         ,  IUnknown,  IUnknArray
          COM_CreateGitUnknArray(),   .
         pCookie   ,  
           0   .

    Remove (DWORD cookie)

         IUnknown,  cookie  .

    Clear ()

           .

    GetUnknown (DWORD cookie, REFIID iid, LPVOID* ppvObject)

         ,  cookie.  iid   
          .     NULL,  IUnknArray 
          COM_CreateGitUnknArray().

    GetCount (DWORD* count)

            .

    IsEmpty ()

         S_FALSE,    ,  - S_OK.


      IEnumerable<CONNECTDATA>
*/

/* {90221A04-CE24-4363-B687-33A32DD544EB} */
static const GUID IID_IUnknArray = {0x90221a04, 0xce24, 0x4363, {0xb6, 0x87, 0x33, 0xa3, 0x2d, 0xd5, 0x44, 0xeb}};

typedef struct _IUnknArrayVtbl  IUnknArrayVtbl;
typedef struct _IUnknArray      IUnknArray;

struct _IUnknArrayVtbl
{
    /* IUnknown */
    STDMETHOD  (QueryInterface  ) (IUnknArray* This, REFIID iid, LPVOID* ppvObject);
    STDMETHOD_ (ULONG, AddRef   ) (IUnknArray* This);
    STDMETHOD_ (ULONG, Release  ) (IUnknArray* This);

    /* IUnknArray */
    STDMETHOD  (SetMaxSize      ) (IUnknArray* This, DWORD count);
    STDMETHOD  (Add             ) (IUnknArray* This, IUnknown* unkn, REFIID iid/*=NULL*/, DWORD* pCookie);
    STDMETHOD  (Remove          ) (IUnknArray* This, DWORD cookie);
    STDMETHOD  (Clear           ) (IUnknArray* This);
    STDMETHOD  (GetUnknown      ) (IUnknArray* This, DWORD cookie, REFIID iid/*=NULL*/, LPVOID* ppvObject);
    STDMETHOD  (GetCount        ) (IUnknArray* This, DWORD* count);
    STDMETHOD  (IsEmpty         ) (IUnknArray* This);
};

struct _IUnknArray
{
    const IUnknArrayVtbl* lpVtbl;
};

#ifdef COBJMACROS
#   define IUnknArray_QueryInterface(This, iid, ppvObject)      ((This)->lpVtbl->QueryInterface (This, iid, ppvObject))
#   define IUnknArray_AddRef(This)                              ((This)->lpVtbl->AddRef (This))
#   define IUnknArray_Release(This)                             ((This)->lpVtbl->Release (This))
#   define IUnknArray_SetMaxSize(This, count)                   ((This)->lpVtbl->SetMaxSize (This, count))
#   define IUnknArray_Add(This, unkn, iid, pCookie)             ((This)->lpVtbl->Add (This, unkn, iid, pCookie))
#   define IUnknArray_Remove(This, cookie)                      ((This)->lpVtbl->Remove (This, cookie))
#   define IUnknArray_Clear(This)                               ((This)->lpVtbl->Clear (This))
#   define IUnknArray_GetUnknown(This, cookie, iid, ppvObject)  ((This)->lpVtbl->GetUnknown (This, cookie, iid, ppvObject))
#   define IUnknArray_GetCount(This, count)                     ((This)->lpVtbl->GetCount (This, count))
#   define IUnknArray_IsEmpty(This)                             ((This)->lpVtbl->IsEmpty (This))
#endif

/* -------------------------------------------------------------------------- */
/*
    CConnectable

       ,   IConnectionPointContainer

     COM_CLASS       
     COM_CLASS_NAME(CConnectable).  COM_OBJECT  
          CConnectable.

    :
        
    HRESULT CreateConnectionPoint (CConnectable* This, REFIID iid, DWORD maxConnections, BOOL useGIT)
    
    :
        This            -  ,   CConnectable
        iid             -  ,   .
        maxConnections  -      .
        useGIT          -  IGlobalInterfaceTable     
                          .   FALSE,   ,  
                             .   
                           TRUE.

      IEnumerable<IConnectionPoint*>
*/

typedef struct _CConnectable CConnectable;

struct _CConnectable
{
    COM_OBJECT object;

    /* implemented interfaces */
    IConnectionPointContainer* container;
    IEnumerable* enumerable;

    /* aggregated interfaces */

    /* private data */
    LPCRITICAL_SECTION lock;
    COM_TREE connectionPoints;

    HRESULT (*CreateConnectionPoint) (CConnectable* This, REFIID iid, DWORD maxConnections);
};

COM_DECLARE_EXTERN_CLASS (CConnectable);

/* -------------------------------------------------------------------------- */
/*                                                              */
/* -------------------------------------------------------------------------- */

/*   */
BOOL                COM_AddItem                 (COM_LIST* list, COM_LIST_ITEM* newItem);
BOOL                COM_InsertItemBefore        (COM_LIST* list, COM_LIST_ITEM* item, COM_LIST_ITEM* newItem);
BOOL                COM_InsertItemAfter         (COM_LIST* list, COM_LIST_ITEM* item, COM_LIST_ITEM* newItem);
void                COM_RemoveItem              (COM_LIST* list, COM_LIST_ITEM* item);
void                COM_ClearList               (COM_LIST* list);

BOOL                COM_IsValidListIterator     (COM_LIST* list, COM_LIST_ITER* iter);
COM_LIST_ITER*      COM_CreateListIterator      (COM_LIST* list);
COM_LIST_ITER*      COM_CopyListIterator        (COM_LIST* list, COM_LIST_ITER* iter);
void                COM_FreeListIterator        (COM_LIST* list, COM_LIST_ITER* iter);
void                COM_InvalidateListIterators (COM_LIST* list, COM_LIST_ITEM* item/*=NULL*/);

/* AVL- */
COM_TREE_NODE*      COM_FirstNode               (COM_TREE* tree);
COM_TREE_NODE*      COM_LastNode                (COM_TREE* tree);
COM_TREE_NODE*      COM_PrevNode                (COM_TREE* tree, COM_TREE_NODE* node);
COM_TREE_NODE*      COM_NextNode                (COM_TREE* tree, COM_TREE_NODE* node);
BOOL                COM_AddNode                 (COM_TREE* tree, LPCVOID key, COM_TREE_NODE* node);
void                COM_RemoveKey               (COM_TREE* tree, LPCVOID key);
COM_TREE_NODE*      COM_FindNode                (COM_TREE* tree, LPCVOID key, COM_TREE_NODE** parent/*=NULL*/);
void                COM_ClearTree               (COM_TREE* tree);

BOOL                COM_IsValidTreeIterator     (COM_TREE* tree, COM_TREE_ITER* iter);
COM_TREE_ITER*      COM_CreateTreeIterator      (COM_TREE* tree);
COM_TREE_ITER*      COM_CopyTreeIterator        (COM_TREE* tree, COM_TREE_ITER* iter);
void                COM_FreeTreeIterator        (COM_TREE* tree, COM_TREE_ITER* iter);
void                COM_InvalidateTreeIterators (COM_TREE* tree, COM_TREE_NODE* node/*=NULL*/);

/*    */
COM_EVENTS*         COM_GetEvents               (void);
void                COM_SetEventsParam          (LPVOID param);

void                COM_GetStats                (COM_STATS* stats);

void                COM_EnumRunningClasses      (int (*callback)(const COM_CLASS*, LPVOID), LPVOID param);
void                COM_EnumRunningObjects      (int (*callback)(COM_OBJECT*, LPVOID), LPVOID param);

/*     GUID */
BOOL                COM_IsEqualGUID             (LPCGUID guid1, LPCGUID guid2);
int                 COM_CompareGUID             (LPCGUID guid1, LPCGUID guid2);
void                COM_CopyGUID                (LPGUID dest, LPCGUID src/*=NULL*/);

IGuid*              COM_GetGuidInterface        (LPCGUID guid);

/*   */
BOOL                COM_Initialize              (void);
void                COM_Cleanup                 (void);

BOOL                COM_CanUnloadNow            (void);

void                COM_EnterEvents             (void);
void                COM_LeaveEvents             (void);

void                COM_EnterObject             (void);
void                COM_LeaveObject             (void);

void                COM_EnterType               (void);
void                COM_LeaveType               (void);

void                COM_LockServer              (BOOL lock);

HRESULT             COM_GetGIT                  (IGlobalInterfaceTable** ppGIT);

/*    */
BOOL                COM_SetDefaultTypeLibrary   (ITypeLib* typeLib);
BOOL                COM_AddTypeLibrary          (ITypeLib* typeLib);
void                COM_ReleaseTypeLibraries    (void);

/*      */
COM_OBJECT*         _COM_GetObject              (LPVOID This);
#define COM_GetObject(type, This)   (type*)_COM_GetObject (This)

HRESULT             COM_CreateObjectInterface   (COM_OBJECT* object, REFIID iid, LPCVOID lpVtbl, LPVOID* ppvObject);
void                COM_DestroyObjectInterface  (LPVOID unkn);
HRESULT             COM_GetInterfaceTypeInfo    (LPVOID unkn, LCID lcid, ITypeInfo** typeInfo);

BOOL                COM_IsAggregatedObject      (COM_OBJECT* object);

LPCRITICAL_SECTION  COM_GetClassLock            (COM_OBJECT* object);
BOOL                COM_IsClassLock             (COM_OBJECT* object, LPCRITICAL_SECTION lock);

/*    */
HRESULT             COM_CreateObject            (const COM_CLASS* _class, IUnknown* pUnkOuter, LPVOID* ppvObject/* COM_OBJECT** */);
HRESULT             COM_CreateInstance          (const COM_CLASS* _class, IUnknown* pUnkOuter, REFIID riid, LPVOID* ppvObject/* IUnknown** */);
HRESULT             COM_GetClassObject          (const COM_CLASS* _class, REFIID riid, LPVOID* ppvObject/* IUnknown** */);

HRESULT             COM_CreateEnumerator        (IEnumerable* container, REFIID enumIID, LPVOID* ppvObject);
HRESULT             COM_CreateUnknArray         (DWORD maxSize, IUnknArray** ppUnknArray);
HRESULT             COM_CreateGitUnknArray      (DWORD maxSize, IUnknArray** ppUnknArray);

/* -------------------------------------------------------------------------- */

#ifdef __cplusplus
};
#endif

#endif /* _rtl_com_h_included_ */
